from typing import Tuple

import pandas as pd

from aeromaps.models.base import AeroMAPSModel


class EnergyIntensity(AeroMAPSModel):
    def __init__(self, name="passenger_energy_intensity", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        energy_per_ask_without_operations_short_range_dropin_fuel: pd.Series,
        energy_per_ask_without_operations_medium_range_dropin_fuel: pd.Series,
        energy_per_ask_without_operations_long_range_dropin_fuel: pd.Series,
        energy_per_rtk_without_operations_freight_dropin_fuel: pd.Series,
        energy_per_ask_without_operations_short_range_hydrogen: pd.Series,
        energy_per_ask_without_operations_medium_range_hydrogen: pd.Series,
        energy_per_ask_without_operations_long_range_hydrogen: pd.Series,
        energy_per_rtk_without_operations_freight_hydrogen: pd.Series,
        energy_per_ask_without_operations_short_range_electric: pd.Series,
        energy_per_ask_without_operations_medium_range_electric: pd.Series,
        energy_per_ask_without_operations_long_range_electric: pd.Series,
        energy_per_rtk_without_operations_freight_electric: pd.Series,
        operations_gain: pd.Series,
        operations_contrails_overconsumption: pd.Series,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Energy consumption per ASK (with operations) calculation using simple models."""

        energy_per_ask_short_range_dropin_fuel = (
            energy_per_ask_without_operations_short_range_dropin_fuel
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_medium_range_dropin_fuel = (
            energy_per_ask_without_operations_medium_range_dropin_fuel
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_long_range_dropin_fuel = (
            energy_per_ask_without_operations_long_range_dropin_fuel
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_rtk_freight_dropin_fuel = (
            energy_per_rtk_without_operations_freight_dropin_fuel
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_short_range_hydrogen = (
            energy_per_ask_without_operations_short_range_hydrogen
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_medium_range_hydrogen = (
            energy_per_ask_without_operations_medium_range_hydrogen
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_long_range_hydrogen = (
            energy_per_ask_without_operations_long_range_hydrogen
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_rtk_freight_hydrogen = (
            energy_per_rtk_without_operations_freight_hydrogen
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )

        energy_per_ask_short_range_electric = (
            energy_per_ask_without_operations_short_range_electric
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_medium_range_electric = (
            energy_per_ask_without_operations_medium_range_electric
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_ask_long_range_electric = (
            energy_per_ask_without_operations_long_range_electric
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )
        energy_per_rtk_freight_electric = (
            energy_per_rtk_without_operations_freight_electric
            * (1 - operations_gain / 100)
            * (1 + operations_contrails_overconsumption / 100)
        )

        self.df.loc[:, "energy_per_ask_short_range_dropin_fuel"] = (
            energy_per_ask_short_range_dropin_fuel
        )
        self.df.loc[:, "energy_per_ask_medium_range_dropin_fuel"] = (
            energy_per_ask_medium_range_dropin_fuel
        )
        self.df.loc[:, "energy_per_ask_long_range_dropin_fuel"] = (
            energy_per_ask_long_range_dropin_fuel
        )
        self.df.loc[:, "energy_per_rtk_freight_dropin_fuel"] = energy_per_rtk_freight_dropin_fuel
        self.df.loc[:, "energy_per_ask_short_range_hydrogen"] = energy_per_ask_short_range_hydrogen
        self.df.loc[:, "energy_per_ask_medium_range_hydrogen"] = (
            energy_per_ask_medium_range_hydrogen
        )
        self.df.loc[:, "energy_per_ask_long_range_hydrogen"] = energy_per_ask_long_range_hydrogen
        self.df.loc[:, "energy_per_rtk_freight_hydrogen"] = energy_per_rtk_freight_hydrogen
        self.df.loc[:, "energy_per_ask_short_range_electric"] = energy_per_ask_short_range_electric
        self.df.loc[:, "energy_per_ask_medium_range_electric"] = (
            energy_per_ask_medium_range_electric
        )
        self.df.loc[:, "energy_per_ask_long_range_electric"] = energy_per_ask_long_range_electric
        self.df.loc[:, "energy_per_rtk_freight_electric"] = energy_per_rtk_freight_electric

        return (
            energy_per_ask_short_range_dropin_fuel,
            energy_per_ask_medium_range_dropin_fuel,
            energy_per_ask_long_range_dropin_fuel,
            energy_per_rtk_freight_dropin_fuel,
            energy_per_ask_short_range_hydrogen,
            energy_per_ask_medium_range_hydrogen,
            energy_per_ask_long_range_hydrogen,
            energy_per_rtk_freight_hydrogen,
            energy_per_ask_short_range_electric,
            energy_per_ask_medium_range_electric,
            energy_per_ask_long_range_electric,
            energy_per_rtk_freight_electric,
        )



# @Time : 04/01/2024 15:21
# @Author : a.salgas
# @File : operational_margin.py
# @Software: PyCharm
import pandas as pd
from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class PassengerAircraftOperationalProfit(AeroMAPSModel):
    def __init__(self, name="passenger_aircraft_operational_profit", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        operational_profit_reference_years: list,
        operational_profit_reference_years_values: list,
    ) -> pd.Series:
        # Simple computation of airline non-operating costs (NOC)

        operational_profit_prospective = AeromapsInterpolationFunction(
            self,
            operational_profit_reference_years,
            operational_profit_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "operational_profit_per_ask"] = operational_profit_prospective
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "operational_profit_per_ask"] = self.df.loc[
                self.prospection_start_year, "operational_profit_per_ask"
            ]
        operational_profit_per_ask = self.df["operational_profit_per_ask"]
        return operational_profit_per_ask



"""
Plots for life cycle assessments interpretation
"""

from pyvis.network import Network
import os
from IPython.display import display, IFrame
from lca_algebraic.activity import Activity

# from aeromaps.models.impacts.life_cycle_assessment.helpers import list_processes
from lca_modeller.helpers import list_processes
import matplotlib
import matplotlib.pyplot as plt

USER_DB = "Foreground DB"
default_process_tree_filename = "process_tree.html"


def process_tree(
    model: Activity,
    foreground_only: bool = True,
    outfile: str = default_process_tree_filename,
    colormap: str = "Pastel2",
):
    """
    Plots an interactive tree to visualize the activities and exchanges declared in the LCA module.
    :param model: The model containing activities and exchanges.
    :param foreground_only: Boolean flag to include only foreground activities.
    :param outfile: Output filename for the generated tree.
    :param colormap: Colormap to use for node coloring.
    """
    # Init network
    net = Network(
        notebook=True, directed=True, layout=True, cdn_resources="remote", filter_menu=False
    )

    # Get processes hierarchy
    df = list_processes(model, foreground_only)
    df["description"] = df["activity"] + "\n" + df["unit"].fillna("")

    # Colors
    unique_values = df["database"].unique()
    value_indices = {value: i for i, value in enumerate(unique_values)}
    norm = matplotlib.colors.Normalize(vmin=0, vmax=len(unique_values), clip=True)
    mapper = plt.cm.ScalarMappable(norm=norm, cmap=plt.get_cmap(colormap))
    colors = df["database"].apply(
        lambda x: matplotlib.colors.to_hex(mapper.to_rgba(value_indices[x]))
    )

    # Populate network
    edge_data = zip(
        df["activity"],
        df["description"],
        df["parents"],
        df["amounts"],
        df["level"],
        df["database"],
        colors,
    )

    edges = {}
    for activity, description, parents, amounts, level, database, color in edge_data:
        src = activity if database == USER_DB else f"{activity}\n[{database}]"
        # desc_array = description.split(',')
        # desc_db = description.split('(')
        # desc = ',\n'.join(desc_array[:2]) if len(desc_array) > 1 else desc_array[0]
        # desc += f"\n({desc_db[-1]}" if len(desc_db) > 1 else ''
        desc = description

        net.add_node(src, desc, title=src, level=level + 1, shape="box", color=color)

        for parent, amount in zip(parents, amounts):
            if parent:
                edge_key = (src, parent)
                if edge_key in edges:  # Multiple edges between the same nodes
                    edges[edge_key].append(str(amount))
                else:  # first time edge links act and its parent: create new node for parent
                    edges[edge_key] = [str(amount)]

    # Add edges to the network with unique labels
    for (src, dst), labels in edges.items():
        if len(labels) == 1:
            net.add_edge(src, dst, label=labels[0], title=f"Exch. 1: {labels[0]}")
        else:
            num_edges = len(labels)
            roundness_step = 0.1 / num_edges  # Adjust the denominator to control spacing
            for i, label in enumerate(labels):
                curve_type = "curvedCW" if i % 2 == 0 else "curvedCCW"
                roundness = roundness_step * (
                    i + 1
                )  # Adjust roundness to avoid excessive curvature
                net.add_edge(
                    src,
                    dst,
                    label=label,
                    title=f"Exch. {i + 1}: {label}",
                    smooth={"enabled": True, "type": curve_type, "roundness": roundness},
                )

    # Set options
    net.set_edge_smooth("vertical")
    net.toggle_physics(False)
    net.show_buttons(filter_=True)
    net.options.layout.hierarchical.nodeSpacing = 200

    # Save
    net.save_graph(outfile)

    # Display in Jupyter Notebook
    display(IFrame(src=os.path.relpath(outfile), width="100%", height=700))



# @Time : 22/02/2023 16:01
# @Author : a.salgas
# @File : biofuel.py
# @Software: PyCharm
from typing import Tuple

import numpy as np
import pandas as pd
from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class BiofuelCost(AeroMAPSModel):
    def __init__(self, name="biofuel_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_biofuel: pd.Series,
        biofuel_hefa_fog_share: pd.Series,
        biofuel_hefa_others_share: pd.Series,
        biofuel_atj_share: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_hefa_fog_emission_factor: pd.Series,
        biofuel_hefa_others_emission_factor: pd.Series,
        biofuel_ft_others_emission_factor: pd.Series,
        biofuel_ft_msw_emission_factor: pd.Series,
        biofuel_atj_emission_factor: pd.Series,
        kerosene_emission_factor: pd.Series,
        kerosene_market_price: pd.Series,
        biofuel_hefa_fog_capex: pd.Series,
        biofuel_hefa_fog_var_opex: pd.Series,
        biofuel_hefa_fog_feedstock_cost: pd.Series,
        biofuel_hefa_others_capex: pd.Series,
        biofuel_hefa_others_var_opex: pd.Series,
        biofuel_hefa_others_feedstock_cost: pd.Series,
        biofuel_hefa_fuel_efficiency: pd.Series,
        biofuel_hefa_oil_efficiency: pd.Series,
        biofuel_ft_others_capex: pd.Series,
        biofuel_ft_others_var_opex: pd.Series,
        biofuel_ft_others_feedstock_cost: pd.Series,
        biofuel_ft_msw_capex: pd.Series,
        biofuel_ft_msw_feedstock_cost: pd.Series,
        biofuel_ft_msw_var_opex: pd.Series,
        biofuel_ft_efficiency: pd.Series,
        biofuel_atj_capex: pd.Series,
        biofuel_atj_var_opex: pd.Series,
        biofuel_atj_feedstock_cost: pd.Series,
        biofuel_atj_efficiency: pd.Series,
        carbon_tax: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        plant_lifespan: float,
        private_discount_rate: float,
        social_discount_rate: float,
        lhv_biofuel: float,
        density_biofuel: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        ### HEFA FOG
        # print('HEFOG')
        (
            plant_building_scenario_hefa_fog,
            plant_building_cost_hefa_fog,
            biofuel_production_hefa_fog,
            carbon_abatement_cost_hefa_fog,
            biofuel_cost_hefa_fog,
            biofuel_mean_capex_share_hefa_fog,
            biofuel_mean_var_opex_share_hefa_fog,
            biofuel_mean_feedstock_share_hefa_fog,
            biofuel_hefa_fog_mfsp,
            biofuel_carbon_tax_hefa_fog,
            biofuel_cost_premium_hefa_fog,
            biofuel_mfsp_carbon_tax_supplement_hefa_fog,
            specific_carbon_abatement_cost_hefa_fog,
            generic_specific_carbon_abatement_cost_hefa_fog,
        ) = self._pathway_computation(
            biofuel_hefa_fog_emission_factor,
            kerosene_emission_factor,
            carbon_tax,
            kerosene_market_price,
            energy_consumption_biofuel,
            biofuel_hefa_fog_share,
            biofuel_hefa_fog_feedstock_cost,
            biofuel_hefa_fog_var_opex,
            biofuel_hefa_fog_capex,
            biofuel_hefa_fuel_efficiency,
            exogenous_carbon_price_trajectory,
            plant_lifespan,
            private_discount_rate,
            social_discount_rate,
            lhv_biofuel,
            density_biofuel,
        )

        self.df.loc[:, "plant_building_scenario_hefa_fog"] = plant_building_scenario_hefa_fog
        self.df.loc[:, "plant_building_cost_hefa_fog"] = plant_building_cost_hefa_fog
        self.df.loc[:, "carbon_abatement_cost_hefa_fog"] = carbon_abatement_cost_hefa_fog
        self.df.loc[:, "specific_carbon_abatement_cost_hefa_fog"] = (
            specific_carbon_abatement_cost_hefa_fog
        )
        self.df.loc[:, "generic_specific_carbon_abatement_cost_hefa_fog"] = (
            generic_specific_carbon_abatement_cost_hefa_fog
        )
        self.df.loc[:, "biofuel_cost_hefa_fog"] = biofuel_cost_hefa_fog
        self.df.loc[:, "biofuel_mean_capex_share_hefa_fog"] = biofuel_mean_capex_share_hefa_fog
        self.df.loc[:, "biofuel_mean_var_opex_share_hefa_fog"] = (
            biofuel_mean_var_opex_share_hefa_fog
        )
        self.df.loc[:, "biofuel_mean_feedstock_share_hefa_fog"] = (
            biofuel_mean_feedstock_share_hefa_fog
        )
        self.df.loc[:, "biofuel_hefa_fog_mfsp"] = biofuel_hefa_fog_mfsp
        self.df.loc[:, "biofuel_carbon_tax_hefa_fog"] = biofuel_carbon_tax_hefa_fog
        self.df.loc[:, "biofuel_cost_premium_hefa_fog"] = biofuel_cost_premium_hefa_fog
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_hefa_fog"] = (
            biofuel_mfsp_carbon_tax_supplement_hefa_fog
        )

        ### HEFA OTHERS
        # print('HEFA OTHER')
        (
            plant_building_scenario_hefa_others,
            plant_building_cost_hefa_others,
            biofuel_production_hefa_others,
            carbon_abatement_cost_hefa_others,
            biofuel_cost_hefa_others,
            biofuel_mean_capex_share_hefa_others,
            biofuel_mean_var_opex_share_hefa_others,
            biofuel_mean_feedstock_share_hefa_others,
            biofuel_hefa_others_mfsp,
            biofuel_carbon_tax_hefa_others,
            biofuel_cost_premium_hefa_others,
            biofuel_mfsp_carbon_tax_supplement_hefa_others,
            specific_carbon_abatement_cost_hefa_others,
            generic_specific_carbon_abatement_cost_hefa_others,
        ) = self._pathway_computation(
            biofuel_hefa_others_emission_factor,
            kerosene_emission_factor,
            carbon_tax,
            kerosene_market_price,
            energy_consumption_biofuel,
            biofuel_hefa_others_share,
            biofuel_hefa_others_feedstock_cost,
            biofuel_hefa_others_var_opex,
            biofuel_hefa_others_capex,
            biofuel_hefa_oil_efficiency * biofuel_hefa_fuel_efficiency,
            exogenous_carbon_price_trajectory,
            plant_lifespan,
            private_discount_rate,
            social_discount_rate,
            lhv_biofuel,
            density_biofuel,
        )

        self.df.loc[:, "plant_building_scenario_hefa_others"] = plant_building_scenario_hefa_others
        self.df.loc[:, "plant_building_cost_hefa_others"] = plant_building_cost_hefa_others
        self.df.loc[:, "carbon_abatement_cost_hefa_others"] = carbon_abatement_cost_hefa_others
        self.df.loc[:, "specific_carbon_abatement_cost_hefa_others"] = (
            specific_carbon_abatement_cost_hefa_others
        )
        self.df.loc[:, "generic_specific_carbon_abatement_cost_hefa_others"] = (
            generic_specific_carbon_abatement_cost_hefa_others
        )
        self.df.loc[:, "biofuel_cost_hefa_others"] = biofuel_cost_hefa_others
        self.df.loc[:, "biofuel_mean_capex_share_hefa_others"] = (
            biofuel_mean_capex_share_hefa_others
        )
        self.df.loc[:, "biofuel_mean_var_opex_share_hefa_others"] = (
            biofuel_mean_var_opex_share_hefa_others
        )
        self.df.loc[:, "biofuel_mean_feedstock_share_hefa_others"] = (
            biofuel_mean_feedstock_share_hefa_others
        )
        self.df.loc[:, "biofuel_hefa_others_mfsp"] = biofuel_hefa_others_mfsp
        self.df.loc[:, "biofuel_carbon_tax_hefa_others"] = biofuel_carbon_tax_hefa_others
        self.df.loc[:, "biofuel_cost_premium_hefa_others"] = biofuel_cost_premium_hefa_others
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_hefa_others"] = (
            biofuel_mfsp_carbon_tax_supplement_hefa_others
        )

        ### FT OTHERS
        # print('FT OTHERS')
        (
            plant_building_scenario_ft_others,
            plant_building_cost_ft_others,
            biofuel_production_ft_others,
            carbon_abatement_cost_ft_others,
            biofuel_cost_ft_others,
            biofuel_mean_capex_share_ft_others,
            biofuel_mean_var_opex_share_ft_others,
            biofuel_mean_feedstock_share_ft_others,
            biofuel_ft_others_mfsp,
            biofuel_carbon_tax_ft_others,
            biofuel_cost_premium_ft_others,
            biofuel_mfsp_carbon_tax_supplement_ft_others,
            specific_carbon_abatement_cost_ft_others,
            generic_specific_carbon_abatement_cost_ft_others,
        ) = self._pathway_computation(
            biofuel_ft_others_emission_factor,
            kerosene_emission_factor,
            carbon_tax,
            kerosene_market_price,
            energy_consumption_biofuel,
            biofuel_ft_others_share,
            biofuel_ft_others_feedstock_cost,
            biofuel_ft_others_var_opex,
            biofuel_ft_others_capex,
            biofuel_ft_efficiency,
            exogenous_carbon_price_trajectory,
            plant_lifespan,
            private_discount_rate,
            social_discount_rate,
            lhv_biofuel,
            density_biofuel,
        )

        self.df.loc[:, "plant_building_scenario_ft_others"] = plant_building_scenario_ft_others
        self.df.loc[:, "plant_building_cost_ft_others"] = plant_building_cost_ft_others
        self.df.loc[:, "carbon_abatement_cost_ft_others"] = carbon_abatement_cost_ft_others
        self.df.loc[:, "specific_carbon_abatement_cost_ft_others"] = (
            specific_carbon_abatement_cost_ft_others
        )
        self.df.loc[:, "generic_specific_carbon_abatement_cost_ft_others"] = (
            generic_specific_carbon_abatement_cost_ft_others
        )
        self.df.loc[:, "biofuel_cost_ft_others"] = biofuel_cost_ft_others
        self.df.loc[:, "biofuel_mean_capex_share_ft_others"] = biofuel_mean_capex_share_ft_others
        self.df.loc[:, "biofuel_mean_var_opex_share_ft_others"] = (
            biofuel_mean_var_opex_share_ft_others
        )
        self.df.loc[:, "biofuel_mean_feedstock_share_ft_others"] = (
            biofuel_mean_feedstock_share_ft_others
        )
        self.df.loc[:, "biofuel_ft_others_mfsp"] = biofuel_ft_others_mfsp
        self.df.loc[:, "biofuel_carbon_tax_ft_others"] = biofuel_carbon_tax_ft_others
        self.df.loc[:, "biofuel_cost_premium_ft_others"] = biofuel_cost_premium_ft_others
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_ft_others"] = (
            biofuel_mfsp_carbon_tax_supplement_ft_others
        )

        ### FT MSW
        # print('FT MSW')
        (
            plant_building_scenario_ft_msw,
            plant_building_cost_ft_msw,
            biofuel_production_ft_msw,
            carbon_abatement_cost_ft_msw,
            biofuel_cost_ft_msw,
            biofuel_mean_capex_share_ft_msw,
            biofuel_mean_var_opex_share_ft_msw,
            biofuel_mean_feedstock_share_ft_msw,
            biofuel_ft_msw_mfsp,
            biofuel_carbon_tax_ft_msw,
            biofuel_cost_premium_ft_msw,
            biofuel_mfsp_carbon_tax_supplement_ft_msw,
            specific_carbon_abatement_cost_ft_msw,
            generic_specific_carbon_abatement_cost_ft_msw,
        ) = self._pathway_computation(
            biofuel_ft_msw_emission_factor,
            kerosene_emission_factor,
            carbon_tax,
            kerosene_market_price,
            energy_consumption_biofuel,
            biofuel_ft_msw_share,
            biofuel_ft_msw_feedstock_cost,
            biofuel_ft_msw_var_opex,
            biofuel_ft_msw_capex,
            biofuel_ft_efficiency,
            exogenous_carbon_price_trajectory,
            plant_lifespan,
            private_discount_rate,
            social_discount_rate,
            lhv_biofuel,
            density_biofuel,
        )

        self.df.loc[:, "plant_building_scenario_ft_msw"] = plant_building_scenario_ft_msw
        self.df.loc[:, "plant_building_cost_ft_msw"] = plant_building_cost_ft_msw
        self.df.loc[:, "carbon_abatement_cost_ft_msw"] = carbon_abatement_cost_ft_msw
        self.df.loc[:, "specific_carbon_abatement_cost_ft_msw"] = (
            specific_carbon_abatement_cost_ft_msw
        )
        self.df.loc[:, "generic_specific_carbon_abatement_cost_ft_msw"] = (
            generic_specific_carbon_abatement_cost_ft_msw
        )
        self.df.loc[:, "biofuel_cost_ft_msw"] = biofuel_cost_ft_msw
        self.df.loc[:, "biofuel_mean_capex_share_ft_msw"] = biofuel_mean_capex_share_ft_msw
        self.df.loc[:, "biofuel_mean_var_opex_share_ft_msw"] = biofuel_mean_var_opex_share_ft_msw
        self.df.loc[:, "biofuel_mean_feedstock_share_ft_msw"] = biofuel_mean_feedstock_share_ft_msw
        self.df.loc[:, "biofuel_ft_msw_mfsp"] = biofuel_ft_msw_mfsp
        self.df.loc[:, "biofuel_carbon_tax_ft_msw"] = biofuel_carbon_tax_ft_msw
        self.df.loc[:, "biofuel_cost_premium_ft_msw"] = biofuel_cost_premium_ft_msw
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_ft_msw"] = (
            biofuel_mfsp_carbon_tax_supplement_ft_msw
        )

        ### ATJ
        # print('ATJ')
        (
            plant_building_scenario_atj,
            plant_building_cost_atj,
            biofuel_production_atj,
            carbon_abatement_cost_atj,
            biofuel_cost_atj,
            biofuel_mean_capex_share_atj,
            biofuel_mean_var_opex_share_atj,
            biofuel_mean_feedstock_share_atj,
            biofuel_atj_mfsp,
            biofuel_carbon_tax_atj,
            biofuel_cost_premium_atj,
            biofuel_mfsp_carbon_tax_supplement_atj,
            specific_carbon_abatement_cost_atj,
            generic_specific_carbon_abatement_cost_atj,
        ) = self._pathway_computation(
            biofuel_atj_emission_factor,
            kerosene_emission_factor,
            carbon_tax,
            kerosene_market_price,
            energy_consumption_biofuel,
            biofuel_atj_share,
            biofuel_atj_feedstock_cost,
            biofuel_atj_var_opex,
            biofuel_atj_capex,
            biofuel_atj_efficiency,
            exogenous_carbon_price_trajectory,
            plant_lifespan,
            private_discount_rate,
            social_discount_rate,
            lhv_biofuel,
            density_biofuel,
        )

        self.df.loc[:, "plant_building_scenario_atj"] = plant_building_scenario_atj
        self.df.loc[:, "plant_building_cost_atj"] = plant_building_cost_atj
        self.df.loc[:, "carbon_abatement_cost_atj"] = carbon_abatement_cost_atj
        self.df.loc[:, "specific_carbon_abatement_cost_atj"] = specific_carbon_abatement_cost_atj
        self.df.loc[:, "generic_specific_carbon_abatement_cost_atj"] = (
            generic_specific_carbon_abatement_cost_atj
        )
        self.df.loc[:, "biofuel_cost_atj"] = biofuel_cost_atj
        self.df.loc[:, "biofuel_mean_capex_share_atj"] = biofuel_mean_capex_share_atj
        self.df.loc[:, "biofuel_mean_var_opex_share_atj"] = biofuel_mean_var_opex_share_atj
        self.df.loc[:, "biofuel_mean_feedstock_share_atj"] = biofuel_mean_feedstock_share_atj
        self.df.loc[:, "biofuel_atj_mfsp"] = biofuel_atj_mfsp
        self.df.loc[:, "biofuel_carbon_tax_atj"] = biofuel_carbon_tax_atj
        self.df.loc[:, "biofuel_cost_premium_atj"] = biofuel_cost_premium_atj
        self.df.loc[:, "biofuel_mfsp_carbon_tax_supplement_atj"] = (
            biofuel_mfsp_carbon_tax_supplement_atj
        )

        biofuel_mean_mfsp = (
            (biofuel_hefa_fog_mfsp * biofuel_hefa_fog_share / 100).fillna(0)
            + (biofuel_hefa_others_mfsp * biofuel_hefa_others_share / 100).fillna(0)
            + (biofuel_ft_others_mfsp * biofuel_ft_others_share / 100).fillna(0)
            + (biofuel_ft_msw_mfsp * biofuel_ft_msw_share / 100).fillna(0)
            + (biofuel_atj_mfsp * biofuel_atj_share / 100).fillna(0)
        )

        self.df.loc[:, "biofuel_mean_mfsp"] = biofuel_mean_mfsp

        for k in range(self.prospection_start_year, self.end_year + 1):
            # check for vals
            valid = []
            if biofuel_atj_share.loc[k] > 0:
                valid.append(biofuel_atj_mfsp.loc[k])
            if biofuel_hefa_fog_share.loc[k] > 0:
                valid.append(biofuel_hefa_fog_mfsp.loc[k])
            if biofuel_ft_others_share.loc[k] > 0:
                valid.append(biofuel_ft_others_mfsp.loc[k])
            if biofuel_ft_msw_mfsp.loc[k] > 0:
                valid.append(biofuel_ft_msw_mfsp.loc[k])
            if biofuel_hefa_others_mfsp.loc[k] > 0:
                valid.append(biofuel_hefa_others_mfsp.loc[k])

            self.df.loc[k, "biofuel_marginal_mfsp"] = np.max(valid)

        biofuel_marginal_mfsp = self.df.loc[:, "biofuel_marginal_mfsp"]
        # MEAN tax
        biofuel_mean_carbon_tax_per_l = (
            (biofuel_mfsp_carbon_tax_supplement_hefa_fog * biofuel_hefa_fog_share / 100).fillna(0)
            + (
                biofuel_mfsp_carbon_tax_supplement_hefa_others * biofuel_hefa_others_share / 100
            ).fillna(0)
            + (biofuel_mfsp_carbon_tax_supplement_ft_others * biofuel_ft_others_share / 100).fillna(
                0
            )
            + (biofuel_mfsp_carbon_tax_supplement_ft_msw * biofuel_ft_msw_share / 100).fillna(0)
            + (biofuel_mfsp_carbon_tax_supplement_atj * biofuel_atj_share / 100).fillna(0)
        )

        self.df.loc[:, "biofuel_mean_carbon_tax_per_l"] = biofuel_mean_carbon_tax_per_l
        return (
            plant_building_scenario_hefa_fog,
            plant_building_cost_hefa_fog,
            carbon_abatement_cost_hefa_fog,
            specific_carbon_abatement_cost_hefa_fog,
            generic_specific_carbon_abatement_cost_hefa_fog,
            biofuel_cost_hefa_fog,
            biofuel_mean_capex_share_hefa_fog,
            biofuel_mean_var_opex_share_hefa_fog,
            biofuel_mean_feedstock_share_hefa_fog,
            biofuel_hefa_fog_mfsp,
            biofuel_carbon_tax_hefa_fog,
            biofuel_cost_premium_hefa_fog,
            biofuel_mfsp_carbon_tax_supplement_hefa_fog,
            plant_building_scenario_hefa_others,
            plant_building_cost_hefa_others,
            carbon_abatement_cost_hefa_others,
            specific_carbon_abatement_cost_hefa_others,
            generic_specific_carbon_abatement_cost_hefa_others,
            biofuel_cost_hefa_others,
            biofuel_mean_capex_share_hefa_others,
            biofuel_mean_var_opex_share_hefa_others,
            biofuel_mean_feedstock_share_hefa_others,
            biofuel_hefa_others_mfsp,
            biofuel_cost_premium_hefa_others,
            biofuel_carbon_tax_hefa_others,
            biofuel_mfsp_carbon_tax_supplement_hefa_others,
            plant_building_scenario_ft_others,
            plant_building_cost_ft_others,
            carbon_abatement_cost_ft_others,
            specific_carbon_abatement_cost_ft_others,
            generic_specific_carbon_abatement_cost_ft_others,
            biofuel_cost_ft_others,
            biofuel_mean_capex_share_ft_others,
            biofuel_mean_var_opex_share_ft_others,
            biofuel_mean_feedstock_share_ft_others,
            biofuel_ft_others_mfsp,
            biofuel_cost_premium_ft_others,
            biofuel_carbon_tax_ft_others,
            biofuel_mfsp_carbon_tax_supplement_ft_others,
            plant_building_scenario_ft_msw,
            plant_building_cost_ft_msw,
            carbon_abatement_cost_ft_msw,
            specific_carbon_abatement_cost_ft_msw,
            generic_specific_carbon_abatement_cost_ft_msw,
            biofuel_cost_ft_msw,
            biofuel_mean_capex_share_ft_msw,
            biofuel_mean_var_opex_share_ft_msw,
            biofuel_mean_feedstock_share_ft_msw,
            biofuel_ft_msw_mfsp,
            biofuel_cost_premium_ft_msw,
            biofuel_carbon_tax_ft_msw,
            biofuel_mfsp_carbon_tax_supplement_ft_msw,
            plant_building_scenario_atj,
            plant_building_cost_atj,
            biofuel_mean_capex_share_atj,
            biofuel_mean_var_opex_share_atj,
            biofuel_mean_feedstock_share_atj,
            biofuel_atj_mfsp,
            carbon_abatement_cost_atj,
            specific_carbon_abatement_cost_atj,
            generic_specific_carbon_abatement_cost_atj,
            biofuel_cost_atj,
            biofuel_cost_premium_atj,
            biofuel_carbon_tax_atj,
            biofuel_mfsp_carbon_tax_supplement_atj,
            biofuel_mean_carbon_tax_per_l,
            biofuel_mean_mfsp,
            biofuel_marginal_mfsp,
        )

    def _pathway_computation(
        self,
        emission_factor: pd.Series,
        kerosene_emission_factor: pd.Series,
        carbon_tax: pd.Series,
        kerosene_market_price: pd.Series,
        energy_consumption_biofuel: pd.Series,
        share: pd.Series,
        biomass_feedstock_cost: pd.Series,
        biofuel_eis_var_opex: pd.Series,
        biofuel_eis_capex: pd.Series,
        plant_eis_efficiency: pd.Series,
        exogenous_carbon_price_trajectory: pd.Series,
        plant_lifespan: float,
        private_discount_rate: float,
        social_discount_rate: float,
        lhv_biofuel: float,
        density_biofuel: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        # Constants:
        construction_time = 3
        load_factor = 0.95

        # Avoided emission factor in gCO2/MJ
        avoided_emission_factor = kerosene_emission_factor - emission_factor

        # Demand scenario for the pathway in MJ
        demand_scenario = energy_consumption_biofuel * share / 100

        indexes = demand_scenario.index

        # Additional plant capacity building needed (in ton/day output capacity) for a given year
        plant_building_scenario = pd.Series(np.zeros(len(indexes)), indexes)
        # Relative CAPEX cost to build the new facilities in M€2020
        plant_building_cost = pd.Series(np.zeros(len(indexes)), indexes)

        # Annual production in MJ
        biofuel_production = pd.Series(np.zeros(len(indexes)), indexes)

        # carbon abatement cost in €/ton
        carbon_abatement_cost = pd.Series(np.zeros(len(indexes)), indexes)

        specific_carbon_abatement_cost = pd.Series(np.nan, indexes)
        generic_specific_carbon_abatement_cost = pd.Series(np.nan, indexes)

        # Total and detailled annual production cost in M€2020
        biofuel_total_cost = pd.Series(np.zeros(len(indexes)), indexes)
        biofuel_capex_cost = pd.Series(np.zeros(len(indexes)), indexes)
        biofuel_opex_cost = pd.Series(np.zeros(len(indexes)), indexes)
        biofuel_feedstock_cost = pd.Series(np.zeros(len(indexes)), indexes)

        # Total extra cost linked to carbon tax in M€2020
        biofuel_carbon_tax_cost = pd.Series(np.zeros(len(indexes)), indexes)
        # Total annual cost premium in M€2020
        biofuel_cost_premium = pd.Series(np.zeros(len(indexes)), indexes)

        # Extra cost on mfsp linked to carbon tax in €/L
        mfsp_supplement_carbon_tax = pd.Series(np.zeros(len(indexes)), indexes)

        # For each year of the demand scenario the demand is matched by the production
        for year in list(demand_scenario.index):
            # Production missing in year n+1 must be supplied by plant built in year n
            if (year + 1) <= self.end_year and biofuel_production[year + 1] < demand_scenario[
                year + 1
            ]:
                # Getting the production not matched by plants already commissioned
                # by creating plants with actual year data technical data
                biofuel_mfsp = self._compute_pathway_year_mfsp(
                    int(construction_time),
                    int(plant_lifespan),
                    year - construction_time,
                    biomass_feedstock_cost,
                    biofuel_eis_var_opex,
                    biofuel_eis_capex,
                    plant_eis_efficiency,
                    private_discount_rate,
                    load_factor,
                    lhv_biofuel,
                    density_biofuel,
                )

                # Getting the production not matched by plants already commissioned [MJ]
                missing_production = demand_scenario[year + 1] - biofuel_production[year + 1]

                # Converting the missing production to a capacity [in kg/day capacity], including availability of plant
                missing_production_kg = missing_production / lhv_biofuel
                missing_production_litres = missing_production_kg / density_biofuel
                capacity_to_build_kg_day = missing_production_kg / load_factor / 365

                capex_year = (
                    capacity_to_build_kg_day * biofuel_eis_capex[year] / 1000000
                )  # conversion in € per ton per day and in M€

                # Adding the new capacity and related costs into the annual cost and capacity series.
                # It is assumed that building costs are equally distributed among the construction years.
                plant_building_scenario[year] = capacity_to_build_kg_day

                for construction_year in range(year - construction_time, year):
                    if self.historic_start_year < construction_year < self.end_year:
                        plant_building_cost[construction_year] += capex_year / construction_time

                # When production ends: either at the end of plant life or the end of the scenario;
                end_bound = int(min(list(demand_scenario.index)[-1], year + plant_lifespan))
                # Adding new plant production to future years and computing total cost associated

                discounted_cumul_cost = 0
                cumul_em = 0
                generic_discounted_cumul_em = 0

                for i in range(year + 1, end_bound + 1):
                    biofuel_total_cost[i] = (
                        biofuel_total_cost[i]
                        + (missing_production_litres * biofuel_mfsp[i]["TOTAL"]) / 1e6
                    )  # €/L and production in litres => /1000000 for M€
                    biofuel_capex_cost[i] = (
                        biofuel_capex_cost[i]
                        + (missing_production_litres * biofuel_mfsp[i]["CAPEX"]) / 1e6
                    )  # M€
                    biofuel_opex_cost[i] = (
                        biofuel_opex_cost[i]
                        + (missing_production_litres * biofuel_mfsp[i]["VAR_OPEX"]) / 1e6
                    )  # M€
                    biofuel_feedstock_cost[i] = (
                        biofuel_feedstock_cost[i]
                        + (missing_production_litres * biofuel_mfsp[i]["FEEDSTOCK"]) / 1e6
                    )  # M€
                    biofuel_production[i] = biofuel_production[i] + missing_production

                for i in range(year, year + int(plant_lifespan)):
                    if i < (self.end_year + 1):
                        discounted_cumul_cost += (
                            biofuel_mfsp[i]["TOTAL"] - kerosene_market_price[i]
                        ) / (1 + social_discount_rate) ** (i - year)
                        cumul_em += (
                            avoided_emission_factor[i] * (lhv_biofuel * density_biofuel) / 1000000
                        )
                        # discounting emissions for non-hotelling scc
                        generic_discounted_cumul_em += (
                            avoided_emission_factor[i]
                            * (lhv_biofuel * density_biofuel)
                            / 1000000
                            * exogenous_carbon_price_trajectory[i]
                            / exogenous_carbon_price_trajectory[year]
                            / (1 + social_discount_rate) ** (i - year)
                        )

                    else:
                        discounted_cumul_cost += (
                            biofuel_mfsp[self.end_year]["TOTAL"]
                            - kerosene_market_price[self.end_year]
                        ) / (1 + social_discount_rate) ** (i - year)
                        cumul_em += (
                            avoided_emission_factor[self.end_year]
                            * (lhv_biofuel * density_biofuel)
                            / 1000000
                        )
                        # discounting emissions for non-hotelling scc, keep last year scc growth rate as future scc growth rate
                        future_scc_growth = (
                            exogenous_carbon_price_trajectory[self.end_year]
                            / exogenous_carbon_price_trajectory[self.end_year - 1]
                        )

                        generic_discounted_cumul_em += (
                            avoided_emission_factor[self.end_year]
                            * (lhv_biofuel * density_biofuel)
                            / 1000000
                            * (
                                exogenous_carbon_price_trajectory[self.end_year]
                                / exogenous_carbon_price_trajectory[year]
                                * (future_scc_growth) ** (i - self.end_year)
                            )
                            / (1 + social_discount_rate) ** (i - year)
                        )

                # Using unitary values for cost and emission possible as long as the plant operates at constant capacity during its life
                # (Volume gets out of cac sums)
                specific_carbon_abatement_cost[year] = discounted_cumul_cost / cumul_em
                generic_specific_carbon_abatement_cost[year] = (
                    discounted_cumul_cost / generic_discounted_cumul_em
                )

            elif (year == self.end_year) or (
                biofuel_production[year + 1] >= demand_scenario[year + 1] > 0
            ):
                specific_carbon_abatement_cost[year] = specific_carbon_abatement_cost[year - 1]
                generic_specific_carbon_abatement_cost[year] = (
                    generic_specific_carbon_abatement_cost[year - 1]
                )

        # MOD -> Scaling down production for diminishing production scenarios.
        # Very weak model, assuming that production not anymore needed by aviation is used elsewhere in the industry.
        # Stranded asset literature could be valuable to model this better.
        # Proportional production scaling

        scaling_factor = demand_scenario / biofuel_production

        if not all(scaling_factor.isna()):
            biofuel_total_cost = biofuel_total_cost * scaling_factor
            biofuel_capex_cost = biofuel_capex_cost * scaling_factor
            biofuel_opex_cost = biofuel_opex_cost * scaling_factor
            biofuel_feedstock_cost = biofuel_feedstock_cost * scaling_factor

        biofuel_mean_mfsp_litre = (
            biofuel_total_cost / (demand_scenario / (lhv_biofuel * density_biofuel)) * 1000000
        )

        biofuel_mean_capex_share = biofuel_capex_cost / biofuel_total_cost * 100
        biofuel_mean_opex_share = biofuel_opex_cost / biofuel_total_cost * 100
        biofuel_mean_feedstock_share = biofuel_feedstock_cost / biofuel_total_cost * 100

        biofuel_cost_premium = (
            (biofuel_mean_mfsp_litre - kerosene_market_price)
            * (demand_scenario / (lhv_biofuel * density_biofuel))
            / 1000000
        )

        # Compute the carbon tax (M€)

        biofuel_carbon_tax_cost = carbon_tax * emission_factor * demand_scenario / 1000000 / 1000000

        mfsp_supplement_carbon_tax = (
            carbon_tax * emission_factor * (lhv_biofuel * density_biofuel) / 1000000
        )

        # Abatement cost in €/tCO2e (= overcost for a ton of biofuel/avoided emissions)
        carbon_abatement_cost = (
            (biofuel_mean_mfsp_litre - kerosene_market_price)
            / (avoided_emission_factor * (lhv_biofuel * density_biofuel))
            * 1000000
        )

        return (
            plant_building_scenario,
            plant_building_cost,
            biofuel_production,
            carbon_abatement_cost,
            biofuel_total_cost,
            biofuel_mean_capex_share,
            biofuel_mean_opex_share,
            biofuel_mean_feedstock_share,
            biofuel_mean_mfsp_litre,
            biofuel_carbon_tax_cost,
            biofuel_cost_premium,
            mfsp_supplement_carbon_tax,
            specific_carbon_abatement_cost,
            generic_specific_carbon_abatement_cost,
        )

    def _compute_pathway_year_mfsp(
        self,
        construction_time,
        plant_lifespan,
        base_year,
        biomass_feedstock_cost,
        biofuel_eis_var_opex,
        biofuel_eis_capex,
        plant_eis_efficiency,
        private_discount_rate,
        load_fact,
        lhv_biofuel,
        density_biofuel,
    ):
        """
        This function computes the MFSP for a given biofuel production pathway for a plant commissioned at the base year.
        Costs are discounted to find a constant MFSP, excepted feedstock, whose price is directly evolving and passed on biofuel MFSP.
        Capex in €/(kg/day capacity)
        Variable opex in € per L of fuel produced (Constant plant entry into service value)
        Feedstock cost in € per MJ of biomass
        Overall plant conversion efficiency
        Biofuel MFSP returned in €/L
        """
        biofuel_prices = {}

        # modification to base year not to use undefined technology costs => either base year or start of prospection year
        technology_year = max(base_year, self.prospection_start_year)

        real_year_days = 365.25 * load_fact
        real_var_opex = biofuel_eis_var_opex[technology_year] * real_year_days
        cap_cost_npv = 0
        var_op_cost_npv = 0
        biofuel_npv = 0

        # Construction of the facility
        for i in range(0, construction_time):
            # The construction is supposed to span over x years, with a uniform cost repartition
            cap_cost_npv += (
                biofuel_eis_capex[technology_year] * density_biofuel / construction_time
            ) / (1 + private_discount_rate) ** i

        # Commissioning of the facility
        for i in range(construction_time, plant_lifespan + construction_time):
            var_op_cost_npv += real_var_opex / (1 + private_discount_rate) ** i
            biofuel_npv += real_year_days / (1 + private_discount_rate) ** i

        cap_cost_lc = cap_cost_npv / biofuel_npv
        var_op_cost_lc = var_op_cost_npv / biofuel_npv

        end_bound = min(
            max(biomass_feedstock_cost.index), base_year + construction_time + plant_lifespan
        )

        for year in range(base_year + construction_time, int(end_bound) + 1):
            feedstock_price = biomass_feedstock_cost[year]
            feedstock_cost = (
                feedstock_price
                * lhv_biofuel
                * density_biofuel
                / plant_eis_efficiency[technology_year]
            )
            biofuel_prices[year] = {
                "TOTAL": cap_cost_lc + var_op_cost_lc + feedstock_cost,
                "CAPEX": cap_cost_lc,
                "VAR_OPEX": var_op_cost_lc,
                "FEEDSTOCK": feedstock_cost,
            }
        return biofuel_prices


class BiofuelCapex(AeroMAPSModel):
    def __init__(self, name="biofuel_capex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        biofuel_hefa_fog_capex_reference_years: list,
        biofuel_hefa_fog_capex_reference_years_values: list,
        biofuel_hefa_others_capex_reference_years: list,
        biofuel_hefa_others_capex_reference_years_values: list,
        biofuel_ft_others_capex_reference_years: list,
        biofuel_ft_others_capex_reference_years_values: list,
        biofuel_ft_msw_capex_reference_years: list,
        biofuel_ft_msw_capex_reference_years_values: list,
        biofuel_atj_capex_reference_years: list,
        biofuel_atj_capex_reference_years_values: list,
        biofuel_hefa_fog_share: pd.Series,
        biofuel_hefa_others_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_atj_share: pd.Series,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biofuel CAPEX (CApital EXPenditures) estimates"""

        # HEFA FOG
        biofuel_hefa_fog_capex = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_fog_capex_reference_years,
            biofuel_hefa_fog_capex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_fog_capex"] = biofuel_hefa_fog_capex

        # HEFA OTHERS
        biofuel_hefa_others_capex = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_others_capex_reference_years,
            biofuel_hefa_others_capex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_others_capex"] = biofuel_hefa_others_capex

        # FT OTHERS
        biofuel_ft_others_capex = AeromapsInterpolationFunction(
            self,
            biofuel_ft_others_capex_reference_years,
            biofuel_ft_others_capex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_others_capex"] = biofuel_ft_others_capex

        # FT MSW
        biofuel_ft_msw_capex = AeromapsInterpolationFunction(
            self,
            biofuel_ft_msw_capex_reference_years,
            biofuel_ft_msw_capex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_msw_capex"] = biofuel_ft_msw_capex

        # ATJ
        biofuel_atj_capex = AeromapsInterpolationFunction(
            self,
            biofuel_atj_capex_reference_years,
            biofuel_atj_capex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_atj_capex"] = biofuel_atj_capex

        # MEAN
        biofuel_mean_capex = (
            biofuel_hefa_fog_capex * biofuel_hefa_fog_share / 100
            + biofuel_hefa_others_capex * biofuel_hefa_others_share / 100
            + biofuel_ft_others_capex * biofuel_ft_others_share / 100
            + biofuel_ft_msw_capex * biofuel_ft_msw_share / 100
            + biofuel_atj_capex * biofuel_atj_share / 100
        )

        self.df.loc[:, "biofuel_mean_capex"] = biofuel_mean_capex

        return (
            biofuel_hefa_fog_capex,
            biofuel_hefa_others_capex,
            biofuel_ft_others_capex,
            biofuel_ft_msw_capex,
            biofuel_atj_capex,
            biofuel_mean_capex,
        )


class BiofuelVarOpex(AeroMAPSModel):
    def __init__(self, name="biofuel_var_opex", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        biofuel_hefa_fog_var_opex_reference_years: list,
        biofuel_hefa_fog_var_opex_reference_years_values: list,
        biofuel_hefa_others_var_opex_reference_years: list,
        biofuel_hefa_others_var_opex_reference_years_values: list,
        biofuel_ft_others_var_opex_reference_years: list,
        biofuel_ft_others_var_opex_reference_years_values: list,
        biofuel_ft_msw_var_opex_reference_years: list,
        biofuel_ft_msw_var_opex_reference_years_values: list,
        biofuel_atj_var_opex_reference_years: list,
        biofuel_atj_var_opex_reference_years_values: list,
        biofuel_hefa_fog_share: pd.Series,
        biofuel_hefa_others_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_atj_share: pd.Series,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biofuel OPEX (Operational EXPenditures) estimates, excepted feedstock energy
        Values should be specified as €/L of fuel produced.
        It is therefore what could be qualified as variable var_opex."""

        # HEFA FOG
        biofuel_hefa_fog_var_opex = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_fog_var_opex_reference_years,
            biofuel_hefa_fog_var_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_fog_var_opex"] = biofuel_hefa_fog_var_opex

        # HEFA OTHERS
        biofuel_hefa_others_var_opex = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_others_var_opex_reference_years,
            biofuel_hefa_others_var_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_others_var_opex"] = biofuel_hefa_others_var_opex

        # FT OTHERS
        biofuel_ft_others_var_opex = AeromapsInterpolationFunction(
            self,
            biofuel_ft_others_var_opex_reference_years,
            biofuel_ft_others_var_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_others_var_opex"] = biofuel_ft_others_var_opex

        # FT MSW
        biofuel_ft_msw_var_opex = AeromapsInterpolationFunction(
            self,
            biofuel_ft_msw_var_opex_reference_years,
            biofuel_ft_msw_var_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_msw_var_opex"] = biofuel_ft_msw_var_opex

        # ATJ
        biofuel_atj_var_opex = AeromapsInterpolationFunction(
            self,
            biofuel_atj_var_opex_reference_years,
            biofuel_atj_var_opex_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_atj_var_opex"] = biofuel_atj_var_opex

        # MEAN
        biofuel_mean_var_opex = (
            biofuel_hefa_fog_var_opex * biofuel_hefa_fog_share / 100
            + biofuel_hefa_others_var_opex * biofuel_hefa_others_share / 100
            + biofuel_ft_others_var_opex * biofuel_ft_others_share / 100
            + biofuel_ft_msw_var_opex * biofuel_ft_msw_share / 100
            + biofuel_atj_var_opex * biofuel_atj_share / 100
        )

        self.df.loc[:, "biofuel_mean_var_opex"] = biofuel_mean_var_opex

        return (
            biofuel_hefa_fog_var_opex,
            biofuel_hefa_others_var_opex,
            biofuel_ft_others_var_opex,
            biofuel_ft_msw_var_opex,
            biofuel_atj_var_opex,
            biofuel_mean_var_opex,
        )


class BiofuelFeedstock(AeroMAPSModel):
    def __init__(self, name="biofuel_feedstock_cost", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        biofuel_hefa_fog_feedstock_cost_reference_years: list,
        biofuel_hefa_fog_feedstock_cost_reference_years_values: list,
        biofuel_hefa_others_feedstock_cost_reference_years: list,
        biofuel_hefa_others_feedstock_cost_reference_years_values: list,
        biofuel_ft_others_feedstock_cost_reference_years: list,
        biofuel_ft_others_feedstock_cost_reference_years_values: list,
        biofuel_ft_msw_feedstock_cost_reference_years: list,
        biofuel_ft_msw_feedstock_cost_reference_years_values: list,
        biofuel_atj_feedstock_cost_reference_years: list,
        biofuel_atj_feedstock_cost_reference_years_values: list,
        biofuel_hefa_fog_share: pd.Series,
        biofuel_hefa_others_share: pd.Series,
        biofuel_ft_others_share: pd.Series,
        biofuel_ft_msw_share: pd.Series,
        biofuel_atj_share: pd.Series,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biofuel feedstock_cost estimates
        Values should be specified as €/MJ of input biomass"""

        # HEFA FOG
        biofuel_hefa_fog_feedstock_cost = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_fog_feedstock_cost_reference_years,
            biofuel_hefa_fog_feedstock_cost_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_fog_feedstock_cost"] = biofuel_hefa_fog_feedstock_cost

        # HEFA OTHERS
        biofuel_hefa_others_feedstock_cost = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_others_feedstock_cost_reference_years,
            biofuel_hefa_others_feedstock_cost_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_others_feedstock_cost"] = biofuel_hefa_others_feedstock_cost

        # FT OTHERS
        biofuel_ft_others_feedstock_cost = AeromapsInterpolationFunction(
            self,
            biofuel_ft_others_feedstock_cost_reference_years,
            biofuel_ft_others_feedstock_cost_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_others_feedstock_cost"] = biofuel_ft_others_feedstock_cost

        # FT MSW
        biofuel_ft_msw_feedstock_cost = AeromapsInterpolationFunction(
            self,
            biofuel_ft_msw_feedstock_cost_reference_years,
            biofuel_ft_msw_feedstock_cost_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_msw_feedstock_cost"] = biofuel_ft_msw_feedstock_cost

        # ATJ
        biofuel_atj_feedstock_cost = AeromapsInterpolationFunction(
            self,
            biofuel_atj_feedstock_cost_reference_years,
            biofuel_atj_feedstock_cost_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_atj_feedstock_cost"] = biofuel_atj_feedstock_cost

        # MEAN
        biofuel_mean_feedstock_cost = (
            biofuel_hefa_fog_feedstock_cost * biofuel_hefa_fog_share / 100
            + biofuel_hefa_others_feedstock_cost * biofuel_hefa_others_share / 100
            + biofuel_ft_others_feedstock_cost * biofuel_ft_others_share / 100
            + biofuel_ft_msw_feedstock_cost * biofuel_ft_msw_share / 100
            + biofuel_atj_feedstock_cost * biofuel_atj_share / 100
        )

        self.df.loc[:, "biofuel_mean_feedstock_cost"] = biofuel_mean_feedstock_cost

        return (
            biofuel_hefa_fog_feedstock_cost,
            biofuel_hefa_others_feedstock_cost,
            biofuel_ft_others_feedstock_cost,
            biofuel_ft_msw_feedstock_cost,
            biofuel_atj_feedstock_cost,
            biofuel_mean_feedstock_cost,
        )



# @Time : 31/05/2024 11:01
# @Author : a.salgas
# @File : electricity_direct_use.py
# @Software: PyCharm
from typing import Tuple

import pandas as pd
from aeromaps.models.base import AeroMAPSModel


class ElectricityDirectUse(AeroMAPSModel):
    def __init__(self, name="electricity_direct_use", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_electric: pd.Series,
        electricity_emission_factor: pd.Series,
        electricity_market_price: pd.Series,
        carbon_tax: pd.Series,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        electricity_direct_use_total_cost = (
            energy_consumption_electric
            * electricity_market_price
            / 3.6  # kWh to MJ
            / 1000000
        )
        self.df.loc[:, "electricity_direct_use_total_cost"] = electricity_direct_use_total_cost

        electricity_direct_use_carbon_tax = (
            energy_consumption_electric
            * electricity_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )

        self.df.loc[:, "electricity_direct_use_carbon_tax"] = electricity_direct_use_carbon_tax

        electricity_direct_use_carbon_tax_kWh = (
            carbon_tax * electricity_emission_factor / 1000000 * 3.6
        )

        self.df.loc[:, "electricity_direct_use_carbon_tax_kWh"] = (
            electricity_direct_use_carbon_tax_kWh
        )

        return (
            electricity_direct_use_total_cost,
            electricity_direct_use_carbon_tax,
            electricity_direct_use_carbon_tax_kWh,
        )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A discipline interfacing a Python function."""

from __future__ import annotations

import ast
import logging
from inspect import getfullargspec
from inspect import getsource
from typing import TYPE_CHECKING
from typing import Callable
from typing import Final
from typing import Union
from typing import get_type_hints

from gemseo.disciplines.auto_py import AutoPyDiscipline
from numpy import array
from numpy import atleast_2d
from numpy import ndarray
from typing_extensions import get_args
from typing_extensions import get_origin

from gemseo.core.discipline.data_processor import DataProcessor
from gemseo.core.discipline import Discipline

from gemseo.utils.data_conversion import split_array_to_dict_of_arrays
from gemseo.utils.source_parsing import get_callable_argument_defaults

from aeromaps.models.base import AeroMAPSModel

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Sequence

DataType = Union[float, ndarray]

LOGGER = logging.getLogger(__name__)


class AutoPyDisciplineBis(Discipline):
    """Wrap a Python function into a discipline.

    A simplified and straightforward way of integrating a discipline
    from a Python function.

    The Python function can take and return only numbers and NumPy arrays.

    The Python function may or may not include default values for input arguments,
    however, if the resulting :class:`.AutoPyDiscipline` is going to be placed inside
    an :class:`.MDF`, a :class:`.BiLevel` formulation or an :class:`.MDA`
    with strong couplings, then the Python function **must** assign default values for
    its input arguments.

    Examples:
        >>> from gemseo.disciplines.auto_py import AutoPyDiscipline
        >>> from numpy import array
        >>> def my_function(x=0., y=0.):
        >>>     z1 = x + 2*y
        >>>     z2 = x + 2*y + 1
        >>>     return z1, z2
        >>>
        >>> discipline = AutoPyDiscipline(my_function)
        >>> discipline.execute()
        {'x': array([0.]), 'y': array([0.]), 'z1': array([0.]), 'z2': array([1.])}
        >>> discipline.execute({"x": array([1.0]), "y": array([-3.2])})
        {'x': array([1.]), 'y': array([-3.2]), 'z1': array([-5.4]), 'z2': array([-4.4])}
    """

    py_func: Callable
    """The Python function to compute the outputs from the inputs."""

    py_jac: Callable | None
    """The Python function to compute the Jacobian from the inputs."""

    # TODO: API: remove since it is not used.
    use_arrays: bool
    """Whether the function is expected to take arrays as inputs and give outputs as
    arrays."""

    # TODO: API: remove since this feature is provided by the base class.
    input_names: list[str]
    """The names of the inputs."""

    # TODO: API: remove since this feature is provided by the base class.
    output_names: list[str]
    """The names of the outputs."""

    data_processor: AutoDiscDataProcessor
    """A data processor forcing input data to float and output data to arrays."""

    sizes: dict[str, int]
    """The sizes of the input and output variables."""

    __LOG_PREFIX: Final[str] = "Discipline %s: py_func has"

    def __init__(
        self,
        py_func: Callable,
        py_jac: Callable | None = None,
        name: str | None = None,
        use_arrays: bool = False,
        grammar_type: Discipline.GrammarType = Discipline.GrammarType.JSON,
    ) -> None:
        """
        Args:
            py_func: The Python function to compute the outputs from the inputs.
            py_jac: The Python function to compute the Jacobian from the inputs;
                its output value must be a 2D NumPy array
                with rows corresponding to the outputs and columns to the inputs.
            name: The name of the discipline.
                If ``None``, use the name of the Python function.
            use_arrays: Whether the function is expected
                to take arrays as inputs and give outputs as arrays.

        Raises:
            TypeError: When ``py_func`` is not callable.
        """  # noqa: D205 D212 D415
        if not callable(py_func):
            msg = "py_func must be callable."
            raise TypeError(msg)

        super().__init__(name=name or py_func.__name__, grammar_type=grammar_type)
        self.py_func = py_func
        self.use_arrays = use_arrays
        self.py_jac = py_jac
        self.input_names = getfullargspec(self.py_func).args
        if "self" in self.input_names:
            self.input_names.remove("self")
        self.output_names = self.__create_output_names()
        have_type_hints = self.__create_grammars()

        if not have_type_hints and not use_arrays:
            # When type hints are used, the conversions will be handled automatically
            # by the grammars.
            self.data_processor = AutoDiscDataProcessor(self.output_names)

        if self.py_jac is None:
            self.set_jacobian_approximation()

        self.__sizes = {}
        self.__jac_shape = []
        self.__input_names_with_namespaces = []
        self.__output_names_with_namespaces = []

    def __create_grammars(self) -> bool:
        """Create the grammars.

        The grammars use type hints from the function if both the arguments and the
        return value have complete type hints. Otherwise, the grammars will have ndarray
        types.

        Returns:
            Whether type hints are used.
        """
        type_hints = get_type_hints(self.py_func)
        return_type = type_hints.pop("return", None)

        # First, determine if both the inputs and outputs have type hints, otherwise
        # that would make things complicated for no good reason.
        names_to_input_types = {}

        if type_hints:
            missing_args_types = set(self.input_names).difference(type_hints.keys())
            if missing_args_types:
                msg = f"{self.__LOG_PREFIX} missing type hints for the arguments: %s."
                LOGGER.warning(msg, self.name, ",".join(missing_args_types))
            else:
                names_to_input_types = type_hints

        names_to_output_types = {}

        if return_type is not None:
            # There could be only one return value of type tuple, or multiple return
            # values that would also be type hinted with tuple.
            if len(self.output_names) == 1:
                names_to_output_types = {self.output_names[0]: return_type}
            else:
                origin = get_origin(return_type)
                if origin is not tuple:
                    msg = (
                        f"{self.__LOG_PREFIX} bad return type hints: "
                        "expecting a tuple of types, got %s."
                    )
                    LOGGER.warning(msg, self.name, return_type)
                else:
                    type_args = get_args(return_type)
                    n_type_args = len(type_args)
                    n_output_names = len(self.output_names)
                    if n_type_args != n_output_names:
                        msg = (
                            f"{self.__LOG_PREFIX} bad return type hints: "
                            "the number of return values and return types shall be "
                            "equal: "
                            "%i return values but %i return type hints."
                        )
                        LOGGER.warning(msg, self.name, n_output_names, n_type_args)
                    else:
                        names_to_output_types = dict(zip(self.output_names, type_args))

        defaults = get_callable_argument_defaults(self.py_func)

        # Second, create the grammar according to the pre-processing above.
        if names_to_input_types and names_to_output_types:
            self.input_grammar.update_from_types(names_to_input_types)
            self.input_grammar.defaults = defaults
            self.output_grammar.update_from_types(names_to_output_types)
            return True

        msg = (
            f"{self.__LOG_PREFIX} inconsistent type hints: "
            "either both the signature arguments and the return values shall have "
            "type hints or none. "
            "The grammars will not use the type hints at all."
        )
        LOGGER.warning(msg, self.name)
        self.input_grammar.update_from_names(self.input_names)

        for key, value in defaults.items():
            if not isinstance(value, ndarray):
                defaults[key] = array([value])
        self.input_grammar.defaults = defaults

        self.output_grammar.update_from_names(self.output_names)
        return False

    def __create_output_names(self) -> list[str]:
        """Create the names of the outputs.

        Returns:
            The names of the outputs.
        """
        output_names = []

        for node in ast.walk(ast.parse(getsource(self.py_func).strip())):
            if not isinstance(node, ast.Return):
                continue

            value = node.value

            if isinstance(value, ast.Tuple):
                temp_output_names = [elt.id for elt in value.elts]
            else:
                temp_output_names = [value.id]

            if output_names and output_names != temp_output_names:
                msg = (
                    "Two return statements use different variable names; "
                    f"{output_names} and {temp_output_names}."
                )
                raise ValueError(msg)

            output_names = temp_output_names

        return output_names

    def _run(self) -> None:
        output_values = self.py_func(**self.get_input_data(with_namespaces=False))
        if len(self.output_names) == 1:
            output_values = {self.output_names[0]: output_values}
        else:
            output_values = dict(zip(self.output_names, output_values))
        self.store_local_data(**output_values)

    def _compute_jacobian(
        self,
        inputs: Iterable[str] | None = None,
        outputs: Iterable[str] | None = None,
    ) -> None:
        """
        Raises:
            RuntimeError: When the analytic Jacobian :attr:`.py_jac` is ``None``.
            ValueError: When the Jacobian shape is inconsistent.
        """  # noqa: D205 D212 D415
        if self.py_jac is None:
            msg = "The analytic Jacobian is missing."
            raise RuntimeError(msg)

        if not self.__sizes:
            for name, value in self._local_data.items():
                if name in self.input_grammar:
                    converter = self.input_grammar.data_converter
                else:
                    converter = self.output_grammar.data_converter
                self.__sizes[name] = converter.get_value_size(name, value)

            in_to_ns = self.input_grammar.to_namespaced
            self.__input_names_with_namespaces = [
                in_to_ns.get(input_name, input_name) for input_name in self.input_names
            ]
            out_to_ns = self.output_grammar.to_namespaced
            self.__output_names_with_namespaces = [
                out_to_ns.get(output_name, output_name) for output_name in self.output_names
            ]
            self.__jac_shape = (
                sum(
                    self.__sizes[output_name] for output_name in self.__output_names_with_namespaces
                ),
                sum(self.__sizes[input_name] for input_name in self.__input_names_with_namespaces),
            )

        func_jac = self.py_jac(**self.get_input_data(with_namespaces=False))
        if len(func_jac.shape) < 2:
            func_jac = atleast_2d(func_jac)
        if func_jac.shape != self.__jac_shape:
            msg = (
                f"The shape {func_jac.shape} "
                "of the Jacobian matrix "
                f"of the discipline {self.name} "
                f"provided by py_jac "
                "does not match "
                f"(output_size, input_size)={self.__jac_shape}."
            )
            raise ValueError(msg)

        self.jac = split_array_to_dict_of_arrays(
            func_jac,
            self.__sizes,
            self.__output_names_with_namespaces,
            self.__input_names_with_namespaces,
        )


class AutoDiscDataProcessor(DataProcessor):
    """A data processor forcing input data to float and output data to arrays.

    Convert all |g| scalar input data to floats, and convert all discipline output data
    to NumPy arrays.
    """

    # TODO: API: this is never used, remove?
    out_names: Sequence[str]
    """The names of the outputs."""

    # TODO: API: this is never used, remove?
    one_output: bool
    """Whether there is a single output."""

    def __init__(
        self,
        out_names: Sequence[str],
    ) -> None:
        """
        Args:
            out_names: The names of the outputs.
        """  # noqa: D205 D212 D415
        super().__init__()
        self.out_names = out_names
        self.one_output = len(out_names) == 1

    def pre_process_data(self, data: dict[str, DataType]) -> dict[str, DataType]:
        """Pre-process the input data.

        Execute a pre-processing of input data
        after they are checked by :meth:`~Discipline.check_input_data`,
        and before the :meth:`~Discipline._run` method of the discipline is called.

        Args:
            data: The data to be processed.

        Returns:
            The processed data
            where one-length NumPy arrays have been replaced with floats.
        """
        processed_data = data.copy()
        for key, val in data.items():
            if len(val) == 1:
                processed_data[key] = float(val[0])

        return processed_data

    def post_process_data(self, data: dict[str, DataType]) -> dict[str, ndarray]:
        """Post-process the output data.

        Execute a post-processing of the output data
        after the :meth:`~Discipline._run` method of the discipline is called,
        and before they are checked by :meth:`~Discipline.check_output_data`.

        Args:
            data: The data to be processed.

        Returns:
            The processed data with NumPy arrays as values.
        """
        processed_data = data.copy()
        for output_name, output_value in processed_data.items():
            if not isinstance(output_value, ndarray):
                processed_data[output_name] = array([output_value])

        return processed_data


# TODO: API: remove since it is not used.
def to_arrays_dict(data: dict[str, DataType]) -> dict[str, ndarray]:
    """Ensure that the values of a dictionary are NumPy arrays.

    Args:
        data: The dictionary whose values must be NumPy arrays.

    Returns:
        The dictionary with NumPy arrays as values.
    """
    for key, value in data.items():
        if not isinstance(value, ndarray):
            data[key] = array([value])
    return data


class AeroMAPSModelWrapper(AutoPyDiscipline):
    def __init__(self, model):
        self.model: AeroMAPSModel = model

        # TODO: Explore possibility to use json grammar
        self.default_grammar_type = Discipline.GrammarType.SIMPLE

        super(AeroMAPSModelWrapper, self).__init__(
            py_func=self.model.compute,
            use_arrays=True,
        )

        if hasattr(self.model, "auto_inputs"):
            self.input_grammar.update_from_types(
                self.model.auto_inputs
            )  # Explicit addition of auto-generated inputs
            if "kwargs" in self.input_grammar.names:
                self.input_grammar.required_names.remove("kwargs")

        if hasattr(self.model, "auto_outputs"):
            self.output_grammar.update_from_types(
                self.model.auto_outputs
            )  # Explicit addition of auto-generated outputs

        self.name = model.__class__.__name__

        self.update_defaults()

    def update_defaults(self):
        for input in self.get_input_data():
            # if self.model.parameters is None:
            #     self.default_inputs[input] = array([0])
            if hasattr(self.model.parameters, input):
                self.default_inputs[input] = getattr(self.model.parameters, input)



"""
Model for Life Cycle Assessment (LCA) of air transportation systems
"""

from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction
import pandas as pd
import numpy as np
import lca_algebraic as agb
from lca_modeller.io.configuration import LCAProblemConfigurator
from typing import Tuple

from aeromaps.core.process import default_parameters_path
from aeromaps.models.parameters import Parameters
from typing import Dict
import xarray as xr

KEY_YEAR = "year"
KEY_METHOD = "method"


class LifeCycleAssessment(AeroMAPSModel):
    def __init__(
        self,
        name: str = "life_cycle_assessment",
        configuration_file: str = None,
        split_by: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(name=name, *args, **kwargs)

        # Get LCA model and LCIA methods
        if configuration_file is None:
            raise ValueError("Configuration file is missing.")
        _, model, methods = LCAProblemConfigurator(configuration_file).generate()
        self.model = model
        self.methods = methods
        self.axis = split_by
        self.params_names = agb.all_params().keys()
        self.params_dict = dict()
        self.xarray_lca = xr.DataArray()

        # Automatically add LCA parameters (except strings) as inputs of this AeroMAPSModel.
        # See gemseo.py for more details about addition of auto-generated inputs.
        self.auto_inputs = dict()
        json_file_parameters = Parameters()
        json_file_parameters.read_json(file_name=default_parameters_path)  # default json file (parameters.json)
        json_parameters_dict = json_file_parameters.to_dict()

        for x in self.params_names:
            # KEY_YEAR is a special parameter treated separately
            if x == KEY_YEAR:
                continue

            # Inputs provided in json file
            elif x in json_parameters_dict.keys():
                value = json_parameters_dict[x]
                self.auto_inputs[x] = type(value)

            # Parameters that should be interpolated from multiple values provided in json file
            # (reference years and corresponding values)
            elif x + "_reference_years" in json_parameters_dict.keys():
                self.auto_inputs[x + "_reference_years"] = type(
                    json_parameters_dict[x + "_reference_years"]
                )
                self.auto_inputs[x + "_reference_years_values"] = type(
                    json_parameters_dict[x + "_reference_years_values"]
                )

            # Parameters that are not in parameters.json are assumed to be outputs from other AeroMAPS models
            else:
                self.auto_inputs[x] = (
                    pd.Series
                )
                # TODO: is their a way to robustify this assumption about data type pd.Series?
                # Maybe by first running the other models and checking the data type of the outputs.

        # Dry run with lca_algebraic to build symbolic expressions of LCIA impacts
        print("Parametrizing LCIA impacts...", end=" ")
        self.lambdas = agb.lca._preMultiLCAAlgebric(self.model, self.methods, axis=self.axis)
        print("Done.")

        # Add the auto-generated outputs to the AeroMAPSModel
        # FIXME: this is a temporary solution. Not defining explicitly the output type cause issue since gemseo is expecting a np.ndarray or pd.Series as output type.
        self.auto_outputs = {"outputs_lca": xr.DataArray}

        # Alternative?
        # This doesn't work for now because the return value of compute() must be a tuple of explicit variables
        # self.auto_outputs = dict()
        # for i, method in enumerate(self.methods):
        #    if self.lambdas[0].axis_keys:
        #        for j, phase in enumerate(self.lambdas[0].axis_keys):
        #            self.auto_outputs[f"ImpactScore_Method_{i}_Axis_{j}"] = pd.Series
        #    else:
        #        self.auto_outputs[f"ImpactScore_Method_{i}"] = pd.Series

    def compute(
        self, **kwargs
    ) -> xr.DataArray:  # Python 3.9+: use builtins tuple instead of Tuple from typing lib
        # Assign values to parameters
        for name in self.params_names:
            # KEY_YEAR is a special parameter treated separately
            if name == KEY_YEAR:
                self.params_dict[name] = list(range(self.prospection_start_year, self.end_year + 1))
                # replace by self.data["years"]["prospective_years"] ?

            # String parameter
            elif isinstance(kwargs.get(name), str):
                self.params_dict[name] = kwargs[name]

            # Single float value
            elif isinstance(kwargs.get(name), float):
                self.params_dict[name] = kwargs[name]

            # Multiple values
            elif isinstance(kwargs.get(name), (list, np.ndarray, pd.Series)):
                param_values = kwargs[name].copy()
                # Check if the length of the parameter values is consistent with the years
                if len(param_values) > self.end_year - self.prospection_start_year + 1:
                    param_values = param_values[
                        -(self.end_year - self.prospection_start_year + 1) :
                    ]
                elif len(param_values) < self.end_year - self.prospection_start_year + 1:
                    raise ValueError(
                        f"Parameter '{name}' has not enough values for the simulation period."
                    )
                self.params_dict[name] = np.nan_to_num(param_values)

            # Parameters that should be interpolated from multiple values provided in json file
            # (reference years and corresponding values)
            elif name + "_reference_years" in kwargs and name + "_reference_years_values" in kwargs:
                param_values = AeromapsInterpolationFunction(
                    self,
                    kwargs[name + "_reference_years"],
                    kwargs[name + "_reference_years_values"],
                    model_name=self.name,
                )
                param_values = param_values.loc[self.prospection_start_year:self.end_year].values
                self.params_dict[name] = np.nan_to_num(param_values)

            # else: the parameter is not provided and will be set to its default value.
            else:
                raise UserWarning(f'Value for LCA parameter "{name}" is not provided. Default value will be used.')

        # Calculate impacts for all parameters at once
        res = self.multiLCAAlgebraicRaw(**self.params_dict)

        # Set the year as the 'x' axis and rename
        res["params"] = self.params_dict[KEY_YEAR]
        res = res.rename({"params": KEY_YEAR})

        # Store xarray to enable user to access data after process calculation
        self.xarray_lca = outputs_lca = res

        # Test: convert xarray to pd.Series to enable connection with other models.
        # Convert each (impact, axis) pair into a pd.Series
        # if self.lambdas[0].axis_keys:
        #    outputs_lca = tuple([
        #        res.sel(impacts=impact, axis=ax).to_series()
        #        for impact in res.coords["impacts"].values
        #        for ax in res.coords["axis"].values
        #    ])
        # else:
        #    outputs_lca = tuple([
        #        res.sel(impacts=impact).to_series() for impact in res.coords["impacts"].values
        #    ])
        # Issue to solve: how to set return with explicit variables? E.g. return tuple(outputs_lca) doesn't work.
        # It has to be for example return (my_var1, my_var2, my_var3, ..., my_var_n) with each variable being a pd.Series or ndarray.

        return outputs_lca

    def multiLCAAlgebraicRaw(self, **params):
        """
        Main parametric LCIA method : Computes LCA by expressing the foreground
        model as symbolic expression of background activities and parameters.
        Then, compute 'static' inventory of the referenced background activities.
        This enables a very fast recomputation of LCA with different parameters,
        useful for stochastic evaluation of parametrized model
        Parameters
        ----------
        models : List of Activities, or Dict of Activities: Lambdas if impacts exprs already calculated
        methods : List of methods, i.e. impacts to consider. Overriden by info from lambdas if provided with models.
        params : Dict[str,ListOrScalar]
                 You should provide named values of all the parameters declared
                 in the models. Values can be single value or list of samples, all of the same size
        axis : keyword to split impacts by phase. Overriden by info from lambdas if provided with models.
        Return
        ------
        lca : 3 dimension xarray of lca results, with dims=("systems", "impacts", "params"]
        """

        models = {self.model: self.lambdas}
        methods = self.methods
        axis = self.axis

        # if isinstance(models, list):
        #    def to_tuple(item):
        #        if isinstance(item, tuple):
        #            return item
        #        else:
        #            return (item, None)
        #    models = dict(to_tuple(item) for item in self.model)

        # elif not isinstance(models, dict):
        #    models = {models: None}

        param_length = agb.params._compute_param_length(params)
        out = np.full((len(models.keys()), len(methods), param_length), np.nan, float)
        axis_keys = None

        for imodel, (model, lambdas) in enumerate(models.items()):
            dbname = model.key[0]
            with agb.DbContext(dbname):
                # Check no params are passed for FixedParams
                for key in params:
                    if key in agb.params._fixed_params():
                        raise ValueError(
                            "Param '%s' is marked as FIXED, but passed in parameters : ignored"
                            % key
                        )

                if not lambdas or len(lambdas) != len(methods):
                    if lambdas:
                        print("Lambdas do not match with len of methods. Recompiling expressions.")
                    lambdas = agb.lca._preMultiLCAAlgebric(model, methods, axis=axis)

                if lambdas[0].has_axis:
                    # Reshape array to add dimension corresponding to axis keys
                    axis_keys = lambdas[0].axis_keys
                    out = np.expand_dims(out, axis=-2)
                    out = np.repeat(out, len(axis_keys), axis=-2)
                for imethod, lambd_with_params in enumerate(lambdas):
                    value = lambd_with_params.compute(**params).value
                    if isinstance(value, dict):
                        # axis values
                        # value = list(float(value[axis]) if axis in value else 0.0 for axis in lambd_with_params.axis_keys)
                        for iaxis, ax in enumerate(axis_keys):
                            out[imodel, imethod, iaxis, :] = np.float_(value[ax])
                    else:
                        out[imodel, imethod, :] = value
        if axis_keys:
            res = xr.DataArray(
                out,
                name="lca",
                coords=[
                    ("systems", np.fromiter((m.key for m in models.keys()), dtype="O")),
                    ("impacts", np.fromiter(methods, dtype="O")),
                    ("axis", np.fromiter(axis_keys, dtype="O")),
                    ("params", list(range(param_length))),
                ],
            )
        else:
            res = xr.DataArray(
                out,
                name="lca",
                coords=[
                    ("systems", np.fromiter((m.key for m in models.keys()), dtype="O")),
                    ("impacts", np.fromiter(methods, dtype="O")),
                    ("params", list(range(param_length))),
                ],
            )
        return res

    def compute_impacts_from_lambdas(
        self,
        **params: Dict[str, agb.SingleOrMultipleFloat],
    ):
        """
        Modified version of compute_impacts from lca_algebraic.
        More like a wrapper of _postLCAAlgebraic, to avoid calling _preLCAAlgebraic which is unecessarily time consuming when lambdas have already been calculated and doesn't have to be updated.
        """
        dfs = dict()

        dbname = self.model.key[0]
        with agb.DbContext(dbname):
            # Check no params are passed for FixedParams
            for key in params:
                if key in agb.params._fixed_params():
                    print("Param '%s' is marked as FIXED, but passed in parameters : ignored" % key)

            # lambdas = _preMultiLCAAlgebric(model, methods, alpha=alpha, axis=axis)  # <-- this is the time-consuming part

            df = agb.lca._postMultiLCAAlgebric(self.methods, self.lambdas, **params)

            model_name = agb.base_utils._actName(self.model)
            while model_name in dfs:
                model_name += "'"

            # param with several values
            list_params = {k: vals for k, vals in params.items() if isinstance(vals, list)}

            # Shapes the output / index according to the axis or multi param entry
            if self.axis:
                df[self.axis] = self.lambdas[0].axis_keys
                df = df.set_index(self.axis)
                df.index.set_names([self.axis])

                # Filter out line with zero output
                df = df.loc[
                    df.apply(
                        lambda row: not (row.name is None and row.values[0] == 0.0),
                        axis=1,
                    )
                ]

                # Rename "None" to others
                df = df.rename(index={None: "_other_"})

                # Sort index
                df.sort_index(inplace=True)

                # Add "total" line
                df.loc["*sum*"] = df.sum(numeric_only=True)

            elif len(list_params) > 0:
                for k, vals in list_params.items():
                    df[k] = vals
                df = df.set_index(list(list_params.keys()))

            else:
                # Single output ? => give the single row the name of the model activity
                df = df.rename(index={0: model_name})

            dfs[model_name] = df

        if len(dfs) == 1:
            df = list(dfs.values())[0]
        else:
            # Concat several dataframes for several models
            df = pd.concat(list(dfs.values()))

        return df



from typing import Tuple

import pandas as pd
from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class HydrogenCostSimple(AeroMAPSModel):
    def __init__(self, name="hydrogen_cost_simple", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_hydrogen: pd.Series,
        liquid_hydrogen_electrolysis_emission_factor: pd.Series,
        liquid_hydrogen_gas_ccs_emission_factor: pd.Series,
        liquid_hydrogen_gas_emission_factor: pd.Series,
        liquid_hydrogen_coal_ccs_emission_factor: pd.Series,
        liquid_hydrogen_coal_emission_factor: pd.Series,
        hydrogen_electrolysis_share: pd.Series,
        hydrogen_gas_ccs_share: pd.Series,
        hydrogen_gas_share: pd.Series,
        hydrogen_coal_ccs_share: pd.Series,
        hydrogen_coal_share: pd.Series,
        electrolysis_h2_mean_mfsp_kg: pd.Series,
        gas_ccs_h2_mean_mfsp_kg: pd.Series,
        gas_h2_mean_mfsp_kg: pd.Series,
        coal_ccs_h2_mean_mfsp_kg: pd.Series,
        coal_h2_mean_mfsp_kg: pd.Series,
        liquefaction_h2_mean_mfsp_kg: pd.Series,
        carbon_tax: pd.Series,
        lhv_hydrogen: float,
        transport_cost_ratio: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        ### Electrolysis ####

        electrolysis_h2_total_cost = (
            electrolysis_h2_mean_mfsp_kg
            / lhv_hydrogen
            * energy_consumption_hydrogen
            * hydrogen_electrolysis_share
            / 100
            / 1e6
        )

        electrolysis_h2_carbon_tax = (
            energy_consumption_hydrogen
            * hydrogen_electrolysis_share
            / 100
            * liquid_hydrogen_electrolysis_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        # M€
        self.df.loc[:, "electrolysis_h2_carbon_tax"] = electrolysis_h2_carbon_tax

        electrolysis_h2_mfsp_carbon_tax_supplement = (
            carbon_tax * liquid_hydrogen_electrolysis_emission_factor / 1000000 * lhv_hydrogen
        )
        # €/kg_H2
        self.df.loc[:, "electrolysis_h2_mfsp_carbon_tax_supplement"] = (
            electrolysis_h2_mfsp_carbon_tax_supplement
        )

        self.df.loc[:, "electrolysis_h2_total_cost"] = electrolysis_h2_total_cost

        ### gas ####

        gas_h2_total_cost = (
            gas_h2_mean_mfsp_kg
            / lhv_hydrogen
            * energy_consumption_hydrogen
            * hydrogen_gas_share
            / 100
            / 1e6
        )
        gas_h2_carbon_tax = (
            energy_consumption_hydrogen
            * hydrogen_gas_share
            / 100
            * liquid_hydrogen_gas_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        # M€
        self.df.loc[:, "gas_h2_carbon_tax"] = gas_h2_carbon_tax

        gas_h2_mfsp_carbon_tax_supplement = (
            carbon_tax * liquid_hydrogen_gas_emission_factor / 1000000 * lhv_hydrogen
        )
        # €/kg_H2
        self.df.loc[:, "gas_h2_mfsp_carbon_tax_supplement"] = gas_h2_mfsp_carbon_tax_supplement

        self.df.loc[:, "gas_h2_total_cost"] = gas_h2_total_cost

        ### gas_ccs ####

        gas_ccs_h2_total_cost = (
            gas_ccs_h2_mean_mfsp_kg
            / lhv_hydrogen
            * energy_consumption_hydrogen
            * hydrogen_gas_ccs_share
            / 100
            / 1e6
        )
        gas_ccs_h2_carbon_tax = (
            energy_consumption_hydrogen
            * hydrogen_gas_ccs_share
            / 100
            * liquid_hydrogen_gas_ccs_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        # M€
        self.df.loc[:, "gas_ccs_h2_carbon_tax"] = gas_ccs_h2_carbon_tax

        gas_ccs_h2_mfsp_carbon_tax_supplement = (
            carbon_tax * liquid_hydrogen_gas_ccs_emission_factor / 1000000 * lhv_hydrogen
        )
        # €/kg_H2
        self.df.loc[:, "gas_ccs_h2_mfsp_carbon_tax_supplement"] = (
            gas_ccs_h2_mfsp_carbon_tax_supplement
        )

        self.df.loc[:, "gas_ccs_h2_total_cost"] = gas_ccs_h2_total_cost

        ### coal ####

        coal_h2_total_cost = (
            coal_h2_mean_mfsp_kg
            / lhv_hydrogen
            * energy_consumption_hydrogen
            * hydrogen_coal_share
            / 100
            / 1e6
        )
        coal_h2_carbon_tax = (
            energy_consumption_hydrogen
            * hydrogen_coal_share
            / 100
            * liquid_hydrogen_coal_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        # M€
        self.df.loc[:, "coal_h2_carbon_tax"] = coal_h2_carbon_tax

        coal_h2_mfsp_carbon_tax_supplement = (
            carbon_tax * liquid_hydrogen_coal_emission_factor / 1000000 * lhv_hydrogen
        )
        # €/kg_H2
        self.df.loc[:, "coal_h2_mfsp_carbon_tax_supplement"] = coal_h2_mfsp_carbon_tax_supplement

        self.df.loc[:, "coal_h2_total_cost"] = coal_h2_total_cost

        ### coal_ccs ####

        coal_ccs_h2_total_cost = (
            coal_ccs_h2_mean_mfsp_kg
            / lhv_hydrogen
            * energy_consumption_hydrogen
            * hydrogen_coal_ccs_share
            / 100
            / 1e6
        )
        coal_ccs_h2_carbon_tax = (
            energy_consumption_hydrogen
            * hydrogen_coal_ccs_share
            / 100
            * liquid_hydrogen_coal_ccs_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        # M€
        self.df.loc[:, "coal_ccs_h2_carbon_tax"] = coal_ccs_h2_carbon_tax

        coal_ccs_h2_mfsp_carbon_tax_supplement = (
            carbon_tax * liquid_hydrogen_coal_ccs_emission_factor / 1000000 * lhv_hydrogen
        )
        # €/kg_H2
        self.df.loc[:, "coal_ccs_h2_mfsp_carbon_tax_supplement"] = (
            coal_ccs_h2_mfsp_carbon_tax_supplement
        )

        self.df.loc[:, "coal_ccs_h2_total_cost"] = coal_ccs_h2_total_cost

        liquefaction_h2_total_cost = (
            liquefaction_h2_mean_mfsp_kg / lhv_hydrogen * energy_consumption_hydrogen / 1e6
        )

        self.df.loc[:, "liquefaction_h2_total_cost"] = liquefaction_h2_total_cost

        transport_h2_total_cost = transport_cost_ratio * (
            electrolysis_h2_total_cost.fillna(0)
            + gas_ccs_h2_total_cost.fillna(0)
            + gas_h2_total_cost.fillna(0)
            + coal_ccs_h2_total_cost.fillna(0)
            + coal_h2_total_cost.fillna(0)
            + liquefaction_h2_total_cost.fillna(0)
        )

        self.df.loc[:, "transport_h2_total_cost"] = transport_h2_total_cost

        transport_h2_cost_per_kg = (
            transport_h2_total_cost / (energy_consumption_hydrogen / lhv_hydrogen) * 1000000
        )

        self.df.loc[:, "transport_h2_cost_per_kg"] = transport_h2_cost_per_kg

        total_hydrogen_supply_cost = (
            electrolysis_h2_total_cost.fillna(0)
            + gas_ccs_h2_total_cost.fillna(0)
            + gas_h2_total_cost.fillna(0)
            + coal_ccs_h2_total_cost.fillna(0)
            + coal_h2_total_cost.fillna(0)
            + liquefaction_h2_total_cost.fillna(0)
            + transport_h2_total_cost.fillna(0)
        )

        self.df.loc[:, "total_hydrogen_supply_cost"] = total_hydrogen_supply_cost
        # M€

        average_hydrogen_mean_mfsp_kg = (
            total_hydrogen_supply_cost / (energy_consumption_hydrogen / lhv_hydrogen) * 1000000
        )
        self.df.loc[:, "average_hydrogen_mean_mfsp_kg"] = average_hydrogen_mean_mfsp_kg
        # €/kg

        average_hydrogen_mean_carbon_tax_kg = (
            (
                electrolysis_h2_carbon_tax.fillna(0)
                + gas_h2_carbon_tax.fillna(0)
                + gas_ccs_h2_carbon_tax.fillna(0)
                + coal_h2_carbon_tax.fillna(0)
                + coal_ccs_h2_carbon_tax.fillna(0)
            )
            / (energy_consumption_hydrogen / lhv_hydrogen)
            * 1000000
        )
        self.df.loc[:, "average_hydrogen_mean_carbon_tax_kg"] = average_hydrogen_mean_carbon_tax_kg

        return (
            electrolysis_h2_total_cost,
            liquefaction_h2_total_cost,
            transport_h2_total_cost,
            total_hydrogen_supply_cost,
            average_hydrogen_mean_mfsp_kg,
            electrolysis_h2_carbon_tax,
            electrolysis_h2_mfsp_carbon_tax_supplement,
            gas_ccs_h2_total_cost,
            gas_h2_total_cost,
            coal_ccs_h2_total_cost,
            coal_h2_total_cost,
            gas_ccs_h2_carbon_tax,
            gas_ccs_h2_mfsp_carbon_tax_supplement,
            gas_h2_carbon_tax,
            gas_h2_mfsp_carbon_tax_supplement,
            coal_ccs_h2_carbon_tax,
            coal_ccs_h2_mfsp_carbon_tax_supplement,
            coal_h2_carbon_tax,
            coal_h2_mfsp_carbon_tax_supplement,
            average_hydrogen_mean_carbon_tax_kg,
            transport_h2_cost_per_kg,
        )


class HydrogenMfspSimple(AeroMAPSModel):
    def __init__(self, name="hydrogen_mfsp_simple", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        gh2_electrolysis_mfsp_simple_reference_years: list,
        gh2_electrolysis_mfsp_simple_reference_years_values: list,
        gh2_gas_ccs_mfsp_simple_reference_years: list,
        gh2_gas_ccs_mfsp_simple_reference_years_values: list,
        gh2_gas_mfsp_simple_reference_years: list,
        gh2_gas_mfsp_simple_reference_years_values: list,
        gh2_coal_ccs_mfsp_simple_reference_years: list,
        gh2_coal_ccs_mfsp_simple_reference_years_values: list,
        gh2_coal_mfsp_simple_reference_years: list,
        gh2_coal_mfsp_simple_reference_years_values: list,
        liquefaction_mfsp_simple_reference_years: list,
        liquefaction_mfsp_simple_reference_years_values: list,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Hydrogen mfsp_simple (Minimal fuel selling price) estimates"""

        # Electrolysis
        electrolysis_h2_mean_mfsp_kg = AeromapsInterpolationFunction(
            self,
            gh2_electrolysis_mfsp_simple_reference_years,
            gh2_electrolysis_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrolysis_h2_mean_mfsp_kg"] = electrolysis_h2_mean_mfsp_kg

        # Gas CCS
        gas_ccs_h2_mean_mfsp_kg = AeromapsInterpolationFunction(
            self,
            gh2_gas_ccs_mfsp_simple_reference_years,
            gh2_gas_ccs_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "gas_ccs_h2_mean_mfsp_kg"] = gas_ccs_h2_mean_mfsp_kg

        # Gas
        gas_h2_mean_mfsp_kg = AeromapsInterpolationFunction(
            self,
            gh2_gas_mfsp_simple_reference_years,
            gh2_gas_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "gas_h2_mean_mfsp_kg"] = gas_h2_mean_mfsp_kg

        # Coal CCS
        coal_ccs_h2_mean_mfsp_kg = AeromapsInterpolationFunction(
            self,
            gh2_coal_ccs_mfsp_simple_reference_years,
            gh2_coal_ccs_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "coal_ccs_h2_mean_mfsp_kg"] = coal_ccs_h2_mean_mfsp_kg

        # Coal
        coal_h2_mean_mfsp_kg = AeromapsInterpolationFunction(
            self,
            gh2_coal_mfsp_simple_reference_years,
            gh2_coal_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "coal_h2_mean_mfsp_kg"] = coal_h2_mean_mfsp_kg

        # Liquefaction
        liquefaction_h2_mean_mfsp_kg = AeromapsInterpolationFunction(
            self,
            liquefaction_mfsp_simple_reference_years,
            liquefaction_mfsp_simple_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "liquefaction_h2_mean_mfsp_kg"] = liquefaction_h2_mean_mfsp_kg

        return (
            electrolysis_h2_mean_mfsp_kg,
            gas_ccs_h2_mean_mfsp_kg,
            gas_h2_mean_mfsp_kg,
            coal_ccs_h2_mean_mfsp_kg,
            coal_h2_mean_mfsp_kg,
            liquefaction_h2_mean_mfsp_kg,
        )



# @Time : 02/02/2024 16:45
# @Author : a.salgas
# @File : recurring_costs.py
# @Software: PyCharm
import pandas as pd

from aeromaps.models.base import AeroMAPSModel


class RecurringCosts(AeroMAPSModel):
    def __init__(self, name="recurring_costs", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        aircraft_in_out_value_dict: dict,
    ) -> dict:
        rc_aircraft_value_dict = {}
        for category, sets in self.fleet_model.fleet.all_aircraft_elements.items():
            # Calculating values of interest for each aircraft
            for aircraft_var in sets:
                # Check if it's a reference aircraft or a normal aircraft...
                if hasattr(aircraft_var, "parameters"):
                    aircraft_var_name = aircraft_var.parameters.full_name
                    rc_cost = aircraft_var.parameters.rc_cost
                else:
                    aircraft_var_name = aircraft_var.full_name
                    rc_cost = aircraft_var.rc_cost

                rc_aircraft_var_name = aircraft_var_name + ":aircraft_recurring_costs"

                # TODO use dictionary if possible once implemented
                # rc_aircraft_value = max(0.0, aircraft_in_out_value_dict[aircraft_var_name] * rc_cost)
                # For now: direct use of fleet model df
                rc_aircraft_value = self.fleet_model.df.loc[
                    :, (aircraft_var_name + ":aircraft_in_out")
                ] * float(rc_cost)
                rc_aircraft_value[rc_aircraft_value < 0] = 0

                self.fleet_model.df = pd.concat(
                    [
                        self.fleet_model.df,
                        rc_aircraft_value.rename(rc_aircraft_var_name),
                    ],
                    axis=1,
                )

                rc_aircraft_value_dict[aircraft_var_name] = rc_aircraft_value
        return rc_aircraft_value_dict



# @Time : 22/02/2023 16:01
# @Author : a.salgas
# @File : biofuel.py
# @Software: PyCharm
from typing import Tuple

import pandas as pd
from aeromaps.models.base import AeroMAPSModel, AeromapsInterpolationFunction


class ElectrofuelCostSimple(AeroMAPSModel):
    def __init__(self, name="electrofuel_cost_simple", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        energy_consumption_electrofuel: pd.Series,
        electrofuel_emission_factor: pd.Series,
        electrofuel_mean_mfsp_litre: pd.Series,
        carbon_tax: pd.Series,
        lhv_electrofuel: float,
        density_electrofuel: float,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        electrofuel_total_cost = (
            electrofuel_mean_mfsp_litre
            / density_electrofuel
            / lhv_electrofuel
            * energy_consumption_electrofuel
            / 1e6
        )
        self.df.loc[:, "electrofuel_total_cost"] = electrofuel_total_cost

        electrofuel_carbon_tax = (
            energy_consumption_electrofuel
            * electrofuel_emission_factor
            / 1000000
            * carbon_tax
            / 1000000
        )
        self.df.loc[:, "electrofuel_carbon_tax"] = electrofuel_carbon_tax

        electrofuel_mfsp_carbon_tax_supplement = (
            carbon_tax
            * electrofuel_emission_factor
            / 1000000
            * lhv_electrofuel
            * density_electrofuel
        )
        self.df.loc[:, "electrofuel_mfsp_carbon_tax_supplement"] = (
            electrofuel_mfsp_carbon_tax_supplement
        )

        return (
            electrofuel_total_cost,
            electrofuel_carbon_tax,
            electrofuel_mfsp_carbon_tax_supplement,
        )


class ElectrofuelMfspSimple(AeroMAPSModel):
    def __init__(self, name="electrofuel_mfsp_simple", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        electrofuel_mfsp_simple_reference_years: list,
        electrofuel_mfsp_simple_reference_years_values: list,
    ) -> pd.Series:
        """Electrofuel MFSP (Minimal fuel selling price) estimates"""

        electrofuel_mean_mfsp_litre = AeromapsInterpolationFunction(
            self,
            electrofuel_mfsp_simple_reference_years,
            electrofuel_mfsp_simple_reference_years_values,
            model_name=self.name,
        )

        self.df.loc[:, "electrofuel_mean_mfsp_litre"] = electrofuel_mean_mfsp_litre

        return electrofuel_mean_mfsp_litre
